from datetime import datetime

from . import models_v2 as _m
from ..base import Client, ModelMixin


class BitfinexPublic(Client, ModelMixin):
    base_url = 'https://api.bitfinex.com/v2/'
    error_keys = ['message']

    def ticker(self, symbol: str):
        data = self.get(f'ticker/{symbol}')
        if self.return_json:
            return data
        return _m.TradingTicker.create_from_json(data)

    def tickers(self, symbols: list):
        data = self.get('tickers', params={
            'symbols': [str(symbol) for symbol in symbols],
        })
        if self.return_json:
            return data
        return {ticker[0]: _m.TradingTicker.create_from_json(ticker[1:])
                for ticker in data}

    def trades(self,
               symbol: str,
               limit: int=None,
               start: float=None,
               end: float=None,
               sort: bool=None):
        if isinstance(start, datetime):
            start = start.timestamp() * 1000
        if isinstance(end, datetime):
            end = end.timestamp() * 1000
        if sort:
            sort = 1 if sort is True else -1
        data = self.get(f'trades/{symbol}/hist', params={
            'limit': limit,
            'start': start,
            'end': end,
            'sort': sort,
        })
        if self.return_json:
            return data
        return [_m.TradingTrade.create_from_json(trade)
                for trade in data]

    def books(self,
              symbol: str,
              precision: str,
              length: int=None):
        data = self.get(f'book/{symbol}/{precision}', params={'len': length})
        if self.return_json:
            return data
        return [_m.TradingBook.create_from_json(book)
                for book in data]

    def stats(self,
              symbol: str,
              key: str,
              size: str,
              side: str,
              section: str,
              sort: bool=None):
        if sort:
            sort = 1 if sort else -1
        data = self.get(
            f'stats1/{key}:{size}:{symbol}:{side}/{section}',
            params={'sort': sort})
        if self.return_json:
            return data
        if section == 'last':
            return _m.Stat.create_from_json(data)
        else:
            return [_m.Stat.create_from_json(stat)
                    for stat in data]

    def stats_last(self,
                   symbol: str,
                   key: str,
                   size: str,
                   side: str,
                   sort: bool=None):
        return self.stats(symbol, key, size, side, 'last', sort)

    def stats_hist(self,
                   symbol: str,
                   key: str,
                   size: str,
                   side: str,
                   sort: bool=None):
        return self.stats(symbol, key, size, side, 'hist', sort)

    def candles(self,
                symbol: str,
                section: str,
                time_frame: str,
                limit: int=None,
                start: float=None,
                end: float=None,
                sort: bool=None):
        if isinstance(start, datetime):
            start = start.timestamp() * 1000
        if isinstance(end, datetime):
            end = end.timestamp() * 1000
        if sort:
            sort = 1 if sort else -1
        data = self.get(
            f'candles/trade:{time_frame}:{symbol}/{section}', params={
                'limit': limit,
                'start': start,
                'end': end,
                'sort': sort,
            }
        )
        if self.return_json:
            return data
        if section == 'last':
            return _m.Candle.create_from_json(data)
        else:
            return [_m.Candle.create_from_json(candle)
                    for candle in data]

    def candles_last(self,
                     symbol: str,
                     time_frame: str,
                     limit: int=None,
                     start: float=None,
                     end: float=None,
                     sort: bool=None):
        return self.candles(
            symbol, 'last', time_frame, limit, start, end, sort)

    def candles_hist(self,
                     symbol: str,
                     time_frame: str,
                     limit: int=None,
                     start: float=None,
                     end: float=None,
                     sort: bool=None):
        return self.candles(
            symbol, 'hist', time_frame, limit, start, end, sort)



from ..base import Client


class CoinMarketCap(Client):
    base_url = 'https://api.coinmarketcap.com/v1/'
    currencies = None
    timeout = 120

    def ticker(self,
               currency: str=None,
               convert: str=None,
               start: int=None,
               limit: int=None):
        params = {
            'start': start,
            'limit': limit,
            'convert': convert,
        }
        if currency:
            if len(currency) == 3:
                currency = self._get_symbol(currency)['value']
            data = self.get(f'ticker/{currency}/', params=params)[0]
        else:
            data = self.get('ticker/', params=params)
        return data

    def price(self,
              currency: str,
              convert: str=None):
        ticker = self.ticker(currency, convert)
        return float(ticker[f"price_{convert or 'usd'}".lower()])

    def stats(self, convert: str=None):
        data = self.get('global/', params={'convert': convert})
        return data

    def _get_currencies(self):
        ticker = self.ticker()
        return {currency['symbol']: dict(value=currency['id'], decimals=8)
                for currency in ticker}

    def _get_symbol(self, currency: str):
        if self.currencies is None:
            self.currencies = self._get_currencies()
        return self.currencies[currency.upper()]



from ..auth import ApiKeyAuth
from ..base import AuthMixin, Client
from ..common import format_date_iso, format_datetime_iso


class OXR(Client, AuthMixin):
    base_url = 'https://openexchangerates.org/api/'
    error_keys = ['error']

    def __init__(self, app_id: str, timeout: int=120, max_retries: int=None):
        super().__init__(timeout, max_retries)
        self.auth = ApiKeyAuth(app_id, api_key_param='app_id')

    def currencies(self):
        """
        Get a JSON list of all currency symbols available from the Open
        Exchange Rates API, along with their full names.
        ref. https://oxr.readme.io/docs/currencies-json
        """
        return self.get('currencies.json')

    def latest(self, 
               base: str=None, 
               symbols: list=None):
        """
        Get latest data.
        ref. https://oxr.readme.io/docs/latest-json
        """
        return self.get('latest.json', base, symbols)

    def historical(self,
                   date_for: str,
                   base: str=None, 
                   symbols: list=None):
        """
        Get daily historical data
        ref. https://oxr.readme.io/docs/historical-json
        """
        date_for = format_date_iso(date_for)
        return self.get(f'historical/{date_for}.json', base, symbols)

    def time_series(self,
                    start: str,
                    end: str,
                    base: str=None, 
                    symbols: list=None):
        """
        Get time-series data.
        ref. https://oxr.readme.io/docs/time-series-json
        """
        return self.get('time-series.json', base, symbols, params={
            'start': format_date_iso(start),
            'end': format_date_iso(end),
        })

    def convert(self,
                value: int,
                from_symbol: str,
                to_symbol: str):
        """
        Convert any money value from one currency to another at the latest
        API rates.
        ref. https://oxr.readme.io/docs/convert
        """
        return self.get(f'convert/{value}/{from_symbol}/{to_symbol}')

    def ohlc(self,
             start_time: str,
             period: str,
             symbols: list=None,
             base: str=None):
        """
        Get historical open, high, low, close (OHLC) and average data.
        ref. https://oxr.readme.io/docs/ohlc-json
        """
        return self.get('ohlc.json', base, symbols, params={
            'start_time': format_datetime_iso(start_time),
            'period': period,
        })

    # OVERRIDES ---------------------------------------------------------------
    def get(self,
            endpoint: str,
            base: str=None,
            symbols: list=None,
            params: dict=None):
        params = params or {}
        if base is not None:
            params['base'] = base
        if isinstance(symbols, list) or isinstance(symbols, tuple):
            symbols = ','.join(symbols)
        if symbols is not None:
            params['symbols'] = symbols
        return super().get(endpoint, params=params)



from ..base import _Enum


# Bitfinex supported symbols
class Symbol(_Enum):
    BCHUSD = 'tBCHUSD'
    BCHBTC = 'tBCHBTC'
    BTCUSD = 'tBTCUSD'
    LTCUSD = 'tLTCUSD'
    ETHUSD = 'tETHUSD'
    ETHBTC = 'tETHBTC'
    ETCUSD = 'tETCUSD'
    ETCBTC = 'tETCBTC'
    BFXUSD = 'tBFXUSD'
    BFXBTC = 'tBFXBTC'
    RRTUSD = 'tRRTUSD'
    RRTBTC = 'tRRTBTC'
    ZECUSD = 'tZECUSD'
    ZECBTC = 'tZECBTC'
    XMRUSD = 'tXMRUSD'
    XMRBTC = 'tXMRBTC'


# Bitfinex supported currencies
class Currency(_Enum):
    BCH = 'bch'
    BTC = 'btc'
    ETH = 'eth'
    USD = 'usd'


# Bitfinex supported precisions
class BookPrecision(_Enum):
    P0 = 'P0'
    P1 = 'P1'
    P2 = 'P2'
    P3 = 'P3'
    R0 = 'R0'



from ..base import Client


class KrakenPublic(Client):
    base_url = 'https://api.kraken.com/0/'
    error_keys = ['error']

    def server_time(self):
        return self.get('public/Time')

    def assets(self):
        return self.get('public/Assets')

    def asset_pairs(self):
        return self.get('public/AssetPairs')

    def ticker(self, symbol: str):
        return self.get('public/Ticker', params={
            'pair': str(symbol),
        })

    def ohcl(self,
             symbol: str,
             interval: int=None,
             since: str=None):
        return self.get('public/OHCL', params={
            'pair': str(symbol),
            'interval': interval,
            'since': since,
        })

    def order_book(self,
                   symbol: str,
                   count: int=None):
        return self.get('public/Depth', params={
            'pair': str(symbol),
            'count': count,
        })

    def trades(self,
               symbol: str,
               since: str=None):
        return self.get('public/Trades', params={
            'pair': str(symbol),
            'since': since,
        })

    def spread(self,
               symbol: str,
               since: str=None):
        return self.get('public/Spread', params={
            'pair': str(symbol),
            'since': since,
        })



from datetime import datetime, timedelta

from ..base import Client
from ..common import current_utc_date, date_range


class CoinDesk(Client):
    base_url = 'http://api.coindesk.com/v1/'
    timeout = 15

    def bpi(self, currency: str):
        return _BPI(self, currency)

    def rate(self, currency: str):
        return _Rate(self, currency)


class _BPI(CoinDesk):

    def __init__(self, parent, currency: str):
        super().__init__(timeout=parent.timeout)
        self.currency = currency.upper()

    def current(self):
        return self.get(f'bpi/currentprice/{self.currency}.json')

    def historical(self,
                   start: datetime,
                   end: datetime=None,
                   include_today: bool=False):
        if isinstance(start, datetime):
            start = start.date()
        if isinstance(end, datetime):
            end = end.date()
        today = current_utc_date()
        end = end or today
        # Validate dates
        assert start <= end, "'start' date must <= 'end' date'!"
        self._validate_historical_date(start)
        self._validate_historical_date(end)
        # If start date is today, return only current BPI
        if start == today:
            current = self.bpi(self.currency).current()
            current['bpi'] = {
                str(today): current['bpi'][self.currency]['rate_float']}
            return current
        # Normal call for historical BPI
        response = self.get('bpi/historical/close.json', params={
            'currency': self.currency,
            'start': start,
            'end': end,
        })
        # If end date is today, add current BPI
        if end == today and include_today:
            response['bpi'][str(today)] = (
                self.rate(self.currency).current())
        # Validate response
        historical_bpi = response['bpi']
        for d in date_range(start, end):
            assert historical_bpi[str(d)], (
                f'{d} is not present in BPI!')

        return response

    @staticmethod
    def _validate_historical_date(date):
        if not date:
            return
        current_date = current_utc_date()
        msg = (f"({date}) must be a date <= current date "
               f"({current_date})")
        assert date <= current_date, msg


class _Rate(CoinDesk):

    def __init__(self, parent, currency: str):
        super().__init__(timeout=parent.timeout)
        self.currency = currency.upper()
        self._bpi = self.bpi(self.currency)

    def current(self):
        response = self._bpi.current()
        rate = response['bpi'][self.currency]['rate_float']
        return rate

    def historical(self,
                   start: datetime,
                   end: datetime=None,
                   include_today: bool=False):
        response = self._bpi.historical(start, end, include_today)
        rate_dict = response['bpi']
        return rate_dict

    def for_date(self, date_for: datetime):
        if isinstance(date_for, datetime):
            date_for = date_for.date()
        rate_dict = self.historical(start=date_for, end=date_for)
        rate = rate_dict[str(date_for)]
        return rate

    def since_date(self,
                   date_since: datetime,
                   include_today: bool=False):
        if isinstance(date_since, datetime):
            date_since = date_since.date()
        rate_dict = self.historical(date_since, None, include_today)
        return rate_dict

    def last_n_days(self,
                    n_days: int,
                    include_today: bool=False):
        start = current_utc_date() - timedelta(days=n_days)
        rate_dict = self.historical(start, None, include_today)
        return rate_dict



from ..base import _Enum, Currency, Market

# Limits
ORDERS_LIMIT = 300
TRANSFERS_LIMIT = 300


class Currency(Currency):
    ARS = dict(value='ARS')
    BCH = dict(value='BCH', decimals=8)
    BTC = dict(value='BTC', decimals=8)
    CLP = dict(value='CLP')
    COP = dict(value='COP')
    ETH = dict(value='ETH', decimals=9)
    LTC = dict(value='LTC', decimals=8)
    PEN = dict(value='PEN')


class Market(Market):
    BCH_ARS = dict(value='BCH-ARS', base=Currency.BCH, quote=Currency.ARS)
    BCH_BTC = dict(value='BCH-BTC', base=Currency.BCH, quote=Currency.BTC)
    BCH_CLP = dict(value='BCH-CLP', base=Currency.BCH, quote=Currency.CLP)
    BCH_COP = dict(value='BCH-COP', base=Currency.BCH, quote=Currency.COP)
    BTC_ARS = dict(value='BTC-ARS', base=Currency.BTC, quote=Currency.ARS)
    BCH_PEN = dict(value='BCH-PEN', base=Currency.BCH, quote=Currency.PEN)
    BTC_CLP = dict(value='BTC-CLP', base=Currency.BTC, quote=Currency.CLP)
    BTC_COP = dict(value='BTC-COP', base=Currency.BTC, quote=Currency.COP)
    BTC_PEN = dict(value='BTC-PEN', base=Currency.BTC, quote=Currency.PEN)
    ETH_ARS = dict(value='ETH-ARS', base=Currency.ETH, quote=Currency.ARS)
    ETH_BTC = dict(value='ETH-BTC', base=Currency.ETH, quote=Currency.BTC)
    ETH_CLP = dict(value='ETH-CLP', base=Currency.ETH, quote=Currency.CLP)
    ETH_COP = dict(value='ETH-COP', base=Currency.ETH, quote=Currency.COP)
    ETH_PEN = dict(value='ETH-PEN', base=Currency.ETH, quote=Currency.PEN)
    LTC_ARS = dict(value='LTC-ARS', base=Currency.LTC, quote=Currency.ARS)
    LTC_BTC = dict(value='LTC-BTC', base=Currency.LTC, quote=Currency.BTC)
    LTC_CLP = dict(value='LTC-CLP', base=Currency.LTC, quote=Currency.CLP)
    LTC_COP = dict(value='LTC-COP', base=Currency.LTC, quote=Currency.COP)
    LTC_PEN = dict(value='LTC-PEN', base=Currency.LTC, quote=Currency.PEN)


class QuotationType(_Enum):
    BID_GIVEN_SIZE = 'bid_given_size'
    BID_GIVEN_EARNED_BASE = 'bid_given_earned_base'
    BID_GIVEN_SPENT_QUOTE = 'bid_given_spent_quote'
    ASK_GIVEN_SIZE = 'ask_given_size'
    ASK_GIVEN_EARNED_QUOTE = 'ask_given_earned_quote'
    ASK_GIVEN_SPENT_BASE = 'ask_given_spent_base'


class OrderType(_Enum):
    ASK = 'Ask'
    BID = 'Bid'


class OrderPriceType(_Enum):
    MARKET = 'market'
    LIMIT = 'limit'


class OrderState(_Enum):
    RECEIVED = 'received'
    PENDING = 'pending'
    TRADED = 'traded'
    CANCELING = 'canceling'
    CANCELED = 'canceled'


class BalanceEvent(_Enum):
    DEPOSIT_CONFIRM = 'deposit_confirm'
    WITHDRAWAL_CONFIRM = 'withdrawal_confirm'
    TRANSACTION = 'transaction'
    TRANSFER_CONFIRMATION = 'transfer_confirmation'


class ReportType(_Enum):
    CANDLESTICK = 'candlestick'
    AVERAGE_PRICES = 'average_prices'



from . import models as _m
from ..base import Client, ModelMixin


class BudaPublic(Client, ModelMixin):
    base_url = 'https://www.buda.com/api/v2/'
    error_keys = ['message']

    def __init__(self,
                 timeout: int=None,
                 host: str=None,
                 return_json: bool=False,
                 max_retries: int=None,
                 backoff_factor: float=None,
                 enable_rate_limit: bool=None):
        # Override base_url
        if host is not None:
            self.base_url = host
        super().__init__(
            timeout, max_retries, backoff_factor, enable_rate_limit,
            return_json=return_json,
        )

    def markets(self):
        data = self.get('markets')
        if self.return_json:
            return data
        return [_m.Market.create_from_json(market)
                for market in data['markets']]

    def market_details(self, market_id: str):
        data = self.get(f'markets/{market_id}')
        if self.return_json:
            return data
        return _m.Market.create_from_json(data['market'])

    def ticker(self, market_id: str):
        data = self.get(f'markets/{market_id}/ticker')
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data['ticker'])

    def order_book(self, market_id: str):
        data = self.get(f'markets/{market_id}/order_book')
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(data['order_book'])

    def trades(self,
               market_id: str,
               timestamp: int=None,
               limit: int=None):
        data = self.get(f'markets/{market_id}/trades', params={
            'timestamp': timestamp,
            'limit': limit,
        })
        if self.return_json:
            return data
        return _m.Trades.create_from_json(data['trades'])



import warnings

from trading_api_wrappers.buda import Buda, BudaAuth, BudaPublic

__all__ = [
    'SURBTC',
]

deprecation_warning = (
    'SurBTC.com has changed to Buda.com, please use the `buda` package.',
    PendingDeprecationWarning)


class SURBTCAuth(BudaAuth):
    def __init__(self, *args, **kwargs):
        warnings.warn(*deprecation_warning)
        super(SURBTCAuth, self).__init__(*args, **kwargs)


class SURBTCPublic(BudaPublic):
    def __init__(self, *args, **kwargs):
        warnings.warn(*deprecation_warning)
        super(SURBTCPublic, self).__init__(*args, **kwargs)


class SURBTC(Buda):
    Auth = SURBTCAuth
    Public = SURBTCPublic

    def __init__(self):
        warnings.warn(*deprecation_warning)



from ..base import _Enum


# Bitfinex supported symbols
class Symbol(_Enum):
    BTCUSD = 'btcusd'
    LTCUSD = 'ltcusd'
    ETHUSD = 'ethusd'
    ETHBTC = 'ethbtc'
    ETCUSD = 'etcusd'
    ETCBTC = 'etcbtc'
    BFXUSD = 'bfxusd'
    BFXBTC = 'bfxbtc'
    RRTUSD = 'rrtusd'
    RRTBTC = 'rrtbtc'
    ZECUSD = 'zecusd'
    ZECBTC = 'zecbtc'
    XMRUSD = 'xmrusd'
    XMRBTC = 'xmrbtc'
    BCHUSD = 'bchusd'
    BCHBTC = 'bchbtc'
    BCHETH = 'bcheth'


# Bitfinex supported currencies
class Currency(_Enum):
    USD = 'usd'
    BTC = 'btc'
    ETH = 'eth'
    BCH = 'bch'
    BFX = 'bfx'



from ..base import _Enum, Market, Currency

# Limits
ORDERS_LIMIT = 100


class Currency(Currency):
    ARS = dict(value='ARS')
    BRL = dict(value='BRL')
    BTC = dict(value='BTC')
    CLP = dict(value='CLP')
    ETH = dict(value='ETH', decimals=9)
    EUR = dict(value='EUR')
    XLM = dict(value='XLM')


class OrderType(_Enum):
    BUY = 'buy'
    SELL = 'sell'


class Market(Market):
    BTC_ARS = dict(value='BTCARS', base=Currency.BTC, quote=Currency.ARS)
    BTC_BRL = dict(value='BTCBRL', base=Currency.BTC, quote=Currency.BRL)
    BTC_CLP = dict(value='BTCCLP', base=Currency.BTC, quote=Currency.CLP)
    BTC_EUR = dict(value='BTCEUR', base=Currency.BTC, quote=Currency.EUR)
    ETH_ARS = dict(value='ETHARS', base=Currency.ETH, quote=Currency.ARS)
    ETH_BRL = dict(value='ETHBRL', base=Currency.ETH, quote=Currency.BRL)
    ETH_CLP = dict(value='ETHCLP', base=Currency.ETH, quote=Currency.CLP)
    ETH_EUR = dict(value='ETHEUR', base=Currency.ETH, quote=Currency.EUR)
    XLM_ARS = dict(value='XLMARS', base=Currency.XLM, quote=Currency.ARS)
    XLM_BRL = dict(value='XLMBRL', base=Currency.XLM, quote=Currency.BRL)
    XLM_CLP = dict(value='XLMCLP', base=Currency.XLM, quote=Currency.CLP)
    XLM_EUR = dict(value='XLMEUR', base=Currency.XLM, quote=Currency.EUR)


class TimeFrame(_Enum):
    DAYS_1 = 1440
    HOURS_1 = 60
    MINUTES_15 = 15
    MINUTES_5 = 5
    MINUTES_1 = 1

    @staticmethod
    def _format_value(value):
        values = {
            1440: 'DAYS_1',
            60: 'HOURS_1',
            15: 'MINUTES_15',
            5: 'MINUTES_5',
            1: 'MINUTES_1',
        }
        return values[value]



from . import constants as _c
from . import models as _m
from .client_auth import BudaAuth
from .client_public import BudaPublic

__all__ = [
    'Buda',
]


class Buda(object):
    # Models
    models = _m
    # Enum Types
    BalanceEvent = _c.BalanceEvent
    Currency = _c.Currency
    Market = _c.Market
    OrderState = _c.OrderState
    OrderType = _c.OrderType
    OrderPriceType = _c.OrderPriceType
    QuotationType = _c.QuotationType
    ReportType = _c.ReportType
    # Clients
    Auth = BudaAuth
    Public = BudaPublic



from collections import namedtuple
from datetime import datetime


def parse_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, '%Y-%m-%d %H:%M')


def parse_iso_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%f')


def check_null(value):
    return value if value != 'null' else None


def int_or_null(value):
    value = check_null(value)
    return int(value) if value else None


def float_or_null(value):
    value = check_null(value)
    return float(value) if value else None


class Pagination(
    namedtuple('pagination', [
        'previous',
        'limit',
        'page',
        'next'
    ])
):
    @classmethod
    def create_from_json(cls, meta):
        if meta:
            return cls(
                previous=int_or_null(meta['previous']),
                limit=int(meta['limit']),
                page=int(meta['page']),
                next=int_or_null(meta['next'])
            )
        return meta


class Ticker(
    namedtuple('ticker', [
        'high',
        'low',
        'ask',
        'bid',
        'last_price',
        'volume',
        'market',
        'timestamp',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, ticker):
        ticker = ticker[0]
        return cls(
            high=float(ticker['high']),
            low=float(ticker['low']),
            ask=float(ticker['ask']),
            bid=float(ticker['bid']),
            last_price=float(ticker['last_price']),
            volume=float(ticker['volume']),
            market=ticker['market'],
            timestamp=parse_iso_datetime(ticker['timestamp']),
            json=ticker,
        )


class OrderBookEntry(
    namedtuple('book_entry', [
        'price',
        'amount',
        'timestamp',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry['price']),
            amount=float(book_entry['amount']),
            timestamp=parse_iso_datetime(book_entry['timestamp']),
            json=book_entry,
        )


class OrderBook(
    namedtuple('order_book', [
        'order_book',
        'pagination',
    ])
):
    @classmethod
    def create_from_json(cls, order_book, pagination):
        return cls(
            order_book=[OrderBookEntry.create_from_json(book_entry)
                        for book_entry in order_book],
            pagination=Pagination.create_from_json(pagination),
        )


class TradesEntry(
    namedtuple('trades_entry', [
        'market_taker',
        'timestamp',
        'price',
        'amount',
        'market',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, trades_entry):
        return cls(
            market_taker=trades_entry['market_taker'],
            timestamp=parse_iso_datetime(trades_entry['timestamp']),
            price=float(trades_entry['price']),
            amount=float(trades_entry['amount']),
            market=trades_entry['market'],
            json=trades_entry,
        )


class Trades(
    namedtuple('trades', [
        'trades',
        'pagination',
    ])
):
    @classmethod
    def create_from_json(cls, trades, pagination):
        return cls(
            trades=[TradesEntry.create_from_json(trades_entry)
                    for trades_entry in trades],
            pagination=Pagination.create_from_json(pagination),
        )


class WalletBalance(
    namedtuple('wallet_balance', [
        'available',
        'balance',
        'wallet',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, balance):
        return cls(
            available=float(balance['available']),
            balance=float(balance['balance']),
            wallet=balance['wallet'],
            json=balance,
        )


class Balance(
    namedtuple('balance', [
        'ARS',
        'CLP',
        'ETH',
    ])
):
    @classmethod
    def create_from_json(cls, balance):
        return cls(
            ARS=cls.get_wallet_balance(balance, 'ARS'),
            CLP=cls.get_wallet_balance(balance, 'CLP'),
            ETH=cls.get_wallet_balance(balance, 'ETH'),
        )

    @staticmethod
    def get_wallet_balance(balance, wallet: str):
        wallet = [b for b in balance if b['wallet'] == wallet]
        return WalletBalance.create_from_json(wallet[0]) if wallet else None


class OrderAmount(
    namedtuple('wallet_balance', [
        'original',
        'remaining',
        'executed',
    ])
):
    @classmethod
    def create_from_json(cls, amount):
        return cls(
            original=float(amount['original']),
            remaining=float_or_null(amount.get('remaining')),
            executed=float_or_null(amount.get('executed')),
        )


class Order(
    namedtuple('order', [
        'id',
        'status',
        'type',
        'price',
        'amount',
        'execution_price',
        'avg_execution_price',
        'market',
        'created_at',
        'updated_at',
        'executed_at',
        'json',
    ])
):
    @classmethod
    def create_from_json(cls, order):
        return cls(
            # Order ID
            id=order['id'],
            # Order status, 'active' or 'executed'
            status=order['status'],
            # Order type, 'buy' or 'sell'
            type=order['type'],
            # Order limit price
            price=float(order['price']),
            # Order amount
            amount=OrderAmount.create_from_json(order['amount']),
            # Order execution price
            execution_price=float_or_null(
                order.get('execution_price')),
            # Order weighted average execution, 0 if not executed
            avg_execution_price=float_or_null(
                order.get('avg_execution_price')),
            # Market pair
            market=order['market'],
            # Order creation timestamp
            created_at=parse_iso_datetime(order['created_at']),
            # Order update timestamp. Only on active orders
            updated_at=parse_iso_datetime(order.get('created_at')),
            # Order execution timestamp. Only on executed orders
            executed_at=parse_iso_datetime(order.get('executed_at')),
            # Order JSON data
            json=order,
        )


class Orders(
    namedtuple('orders', [
        'orders',
        'pagination',
    ])
):
    @classmethod
    def create_from_json(cls, orders, pagination):
        return cls(
            orders=[Order.create_from_json(order) for order in orders],
            pagination=Pagination.create_from_json(pagination),
        )



from requests import PreparedRequest as P

from .client_public import BitstampPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class BitstampHMACAuth(HMACAuth):

    signature_delimiter = ''

    def __init__(self,
                 api_key: str,
                 secret: str,
                 customer_id: (str, int),
                 **kwargs):
        super().__init__(api_key, secret, **kwargs)
        self.customer_id = str(customer_id)

    def add_api_key(self, r: P):
        r.body['key'] = self.api_key

    def add_nonce(self, r: P, nonce: str):
        r.body['nonce'] = nonce

    def add_signature(self, r: P, nonce: str):
        message = self.build_message(r, nonce)
        signature = self.sign(message)
        r.body['signature'] = signature.upper()

    def auth_request(self, r: P, nonce: str):
        r.body = self.parse_data(r.body)
        super().auth_request(r, nonce)
        r.prepare_body(data=r.body, files=None)

    def build_message(self, r: P, nonce: str):
        components = [nonce, self.customer_id, self.api_key]
        message = self.signature_delimiter.join(components)
        return message


class BitstampAuth(BitstampPublic, AuthMixin):
    auth_cls = BitstampHMACAuth

    def __init__(self,
                 key: str,
                 secret: str,
                 customer_id: (str, int),
                 timeout: int=None,
                 max_retries: int=None,
                 backoff_factor: float=None,
                 enable_rate_limit: bool=None):
        super().__init__(timeout, max_retries, backoff_factor,
                         enable_rate_limit)
        self.add_auth(key, secret, customer_id)

    # Private user data -------------------------------------------------------
    def account_balance(self, currency_pair: str=None):
        """
        Returns amounts available and reserved per currency,
        also fee percentage for the currency pair, or for all currency pairs
        if no pair was specified:

            {'btc_reserved': '0',
             'fee': '0.5000',
             'btc_available': '2.30856098',
             'usd_reserved': '0',
             'btc_balance': '2.30856098',
             'usd_balance': '114.64',
             'usd_available': '114.64',
             ...
             ---If currency pair was specified:
             'fee': '',
             ---If currency pair was not specified:
             'btcusd_fee': '0.25',
             'btceur_fee': '0.25',
             'eurusd_fee': '0.20',
             ...
             }
        """
        endpoint = self._endpoint_for(
            f'balance/{currency_pair}' if currency_pair else 'balance')
        return self.post(endpoint)

    def user_transactions(self,
                          currency_pair: str=None,
                          offset: int=None,
                          limit: int=None,
                          sort_desc: bool=None):
        """
        Returns descending list of transactions. Every transaction (dictionary)
        contains:
        
            {'usd': '-39.25',
             'datetime': '2013-03-26 18:49:13',
             'fee': '0.20',
             'btc': '0.50000000',
             'type': 2,
             'id': 213642}
             
        Instead of the keys btc and usd, it can contain other currency codes
        """
        payload = {
            'offset': offset,
            'limit': limit,
        }
        if sort_desc is not None:
            payload['sort'] = 'desc' if sort_desc else 'asc'
        endpoint = self._endpoint_for('user_transactions')
        if currency_pair:
            endpoint = f'{endpoint}/{currency_pair}/'
        return self.post(endpoint, data=payload)

    # Orders ------------------------------------------------------------------
    def open_orders(self, currency_pair: str=None):
        """
        Returns JSON list of open orders. Each order is represented as a
        dictionary.
        """
        endpoint = self._endpoint_for(f"open_orders/{currency_pair or 'all'}")
        return self.post(endpoint)

    def orders_status(self, order_id: int):
        """
        Returns dictionary.
        - status: 'Finished'
          or      'In Queue'
          or      'Open'
        - transactions: list of transactions
          Each transaction is a dictionary with the following keys:
              btc, usd, price, type, fee, datetime, tid
          or  btc, eur, ....
          or  eur, usd, ....
        """
        endpoint = self._endpoint_for('order_status', version=1)
        return self.post(endpoint, data={'id': order_id})

    def cancel_order(self, order_id: int):
        """
        Cancel the order specified by order_id.

        Returns dictionary of the canceled order, containing the keys:
        id, type, price, amount
        """
        endpoint = self._endpoint_for('cancel_order')
        return self.post(endpoint, data={'id': order_id})

    def cancel_all_orders(self):
        """
        Cancel all open orders.

        Returns True if it was successful
        """
        endpoint = self._endpoint_for('cancel_all_orders', version=1)
        return self.post(endpoint)

    def _limit_order(self,
                     endpoint,
                     currency_pair: str,
                     amount: float,
                     price: float,
                     limit_price: float,
                     daily_order: bool=None):
        endpoint = self._endpoint_for(endpoint % currency_pair)
        return self.post(endpoint, data={
            'amount': amount,
            'price': price,
            'limit_price': limit_price,
            'daily_order': daily_order,
        })

    def _market_order(self,
                      endpoint: str,
                      currency_pair: str,
                      amount: float):
        endpoint = self._endpoint_for(endpoint % currency_pair)
        return self.post(endpoint, data={'amount': amount})

    def buy_limit_order(self,
                        currency_pair: str,
                        amount: float,
                        price: float,
                        limit_price: float,
                        daily_order: bool=None):
        """
        Order to buy amount for specified price.
        """
        return self._limit_order('buy/%s', currency_pair, amount,
                                 price, limit_price, daily_order)

    def sell_limit_order(self,
                         currency_pair: str,
                         amount: float,
                         price: float,
                         limit_price: float,
                         daily_order: bool=None):
        """
        Order to sell amount for specified price.
        """
        return self._limit_order('sell/%s', currency_pair, amount,
                                 price, limit_price, daily_order)

    def buy_market_order(self,
                         currency_pair: str,
                         amount: float):
        """
        Order to buy amount for market price.
        """
        return self._market_order('buy/market/%s', currency_pair, amount)

    def sell_market_order(self,
                          currency_pair: str,
                          amount: float):
        """
        Order to sell amount for market price.
        """
        return self._market_order('sell/market/%s', currency_pair, amount)

    # Withdrawals -------------------------------------------------------------
    def withdrawal_requests(self, time_delta: int=None):
        """
        Returns list of withdrawal requests.

        Each request is represented as a dictionary.

        By default, the last 24 hours (86400 seconds) are returned.
        """
        endpoint = self._endpoint_for('withdrawal-requests')
        return self.post(endpoint, data={'timedelta': time_delta})

    def withdrawal(self,
                   method: str,
                   address: str,
                   amount: float,
                   version: int=2,
                   **kwargs):
        endpoint = self._endpoint_for(f'{method}_withdrawal', version)
        return self.post(endpoint, data={
            'amount': amount,
            'address': address,
            **kwargs
        })

    def bitcoin_withdrawal(self, address: str, amount: float,
                           instant: bool=None):
        """
        Send bitcoin to another bitcoin wallet specified by address.
        """
        return self.withdrawal('bitcoin', address, amount, **{
            'version': 1,
            'instant': (1 if instant else 0) if instant else None,
        })

    def ripple_withdrawal(self, address: str, amount: float, currency: str):
        """
        Returns true if successful.
        """
        return self.withdrawal('ripple', address, amount,
                               version=1, currency=currency)

    def litecoin_withdrawal(self, address: str, amount: float):
        """
        Send litecoin to another litecoin wallet specified by address.
        """
        return self.withdrawal('ltc', address, amount)

    def eth_withdrawal(self, address: str, amount: float):
        """
        Send ether to another ether wallet specified by address.
        """
        return self.withdrawal('eth', address, amount)

    def bch_withdrawal(self, address: str, amount: float):
        """
        Send bitcoin cash to another bitcoin cash wallet specified by address.
        """
        return self.withdrawal('bch', address, amount)

    def xrp_withdrawal(self, address: str, amount: float,
                       destination_tag: str=None):
        """
        Sends xrp to another xrp wallet specified by address.
        Returns withdrawal id.
        """
        return self.withdrawal('xrp', address, amount,
                               destination_tag=destination_tag)

    # Deposits ----------------------------------------------------------------
    def deposit_address(self,
                        method: str,
                        version: int=2):
        endpoint = self._endpoint_for(f'{method}_address', version)
        return self.post(endpoint)

    def bitcoin_deposit_address(self):
        """
        Returns bitcoin deposit address as unicode string
        """
        return self.deposit_address('bitcoin_deposit', version=1)

    def ripple_deposit_address(self):
        """
        Returns ripple deposit address as unicode string.
        """
        return self.deposit_address('ripple', version=1)

    def litecoin_deposit_address(self):
        """
        Returns litecoin deposit address as unicode string
        """
        return self.deposit_address('ltc')

    def eth_deposit_address(self):
        """
        Returns ethereum deposit address as unicode string
        """
        return self.deposit_address('eth')

    def bch_deposit_address(self):
        """
        Returns bitcoin cash deposit address as unicode string
        """
        return self.deposit_address('bch')

    def xrp_deposit_address(self):
        """
        Returns ripple deposit address and destination tag as dictionary.
        Example:
            {'destination_tag': 53965834,
             'address': 'rDsBeAmAa4FFwbQTJp9Rs84Q56vCiWCaBx'}
        """
        return self.deposit_address('xrp')

    def unconfirmed_bitcoin_deposits(self):
        """
        Returns JSON list of unconfirmed bitcoin transactions.
        Each transaction is represented as dictionary:

        amount
          bitcoin amount
        address
          deposit address used
        confirmations
          number of confirmations
        """
        endpoint = self._endpoint_for('unconfirmed_btc', version=1)
        return self.post(endpoint)

    # Transfers ---------------------------------------------------------------
    def _transfer(self,
                  endpoint: str,
                  amount: float,
                  currency: str,
                  sub_account: int=None):
        endpoint = self._endpoint_for(endpoint)
        return self.post(endpoint, data={
            'amount': amount,
            'currency': currency,
            'subAccount': sub_account,
        })

    def transfer_to_main(self,
                         amount: float,
                         currency: str,
                         sub_account: int=None):
        """
        Returns dictionary with status.
        sub account has to be the numerical id of the sub account, not the name
        """
        return self._transfer('transfer-to-main',
                              amount, currency, sub_account)

    def transfer_from_main(self,
                           amount: float,
                           currency: str,
                           sub_account: int):
        """
        Returns dictionary with status.
        sub account has to be the numerical id of the sub account, not the name
        """
        return self._transfer('transfer-from-main',
                              amount, currency, sub_account)

    # TODO: Bank methods



import base64

from requests import PreparedRequest as P

from .client_public_v1 import BitfinexPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class BitfinexHMACAuth(HMACAuth):

    api_key_header = 'x-bfx-apikey'
    signature_header = 'x-bfx-signature'
    payload_header = 'x-bfx-payload'
    algorithm = 'sha384'

    def add_nonce(self, r: P, nonce: str):
        pass

    def add_signature(self, r: P, nonce: str):
        message = self.build_message(r, nonce)
        r.headers[self.payload_header] = message
        signature = self.sign(message)
        r.headers[self.signature_header] = signature

    def build_message(self, r: P, nonce: str):
        body = self.load_json(r.body)
        body['request'] = r.path_url
        body['nonce'] = nonce
        encoded_msg = base64.b64encode(self.encode_json(body))
        return encoded_msg


class BitfinexAuth(BitfinexPublic, AuthMixin):
    auth_cls = BitfinexHMACAuth

    def __init__(self,
                 key: str,
                 secret: str,
                 timeout: int=None,
                 max_retries: int=None,
                 backoff_factor: float=None,
                 enable_rate_limit: bool=None):
        super().__init__(timeout, max_retries, backoff_factor,
                         enable_rate_limit)
        self.add_auth(key, secret)

    # INFO --------------------------------------------------------------------
    # Return information about your account (trading fees).
    def account_info(self):
        return self.get('account_infos')

    # Return information about your account (trading fees).
    def summary(self):
        return self.get('summary')

    # Check the permissions of the key being used to generate this request.
    def key_info(self):
        return self.get('key_info')

    # See your trading wallet information for margin trading.
    def margin_info(self):
        return self.get('margin_infos')

    # See your balances.
    def balances(self):
        return self.get('balances')

    # MOVEMENTS ---------------------------------------------------------------
    # Return your deposit address to make a new deposit.
    def new_deposit(self,
                    method: str,
                    wallet_name: str,
                    renew: bool=None):
        return self.post('deposit/new', json={
            'method': method,
            'wallet_name': wallet_name,
            'renew': renew if renew is not None else None,
        })

    # Allow you to move available balances between your wallets.
    def transfer(self,
                 amount: float,
                 currency: str,
                 wallet_from: str,
                 wallet_to: str):
        return self.post('transfer', json={
            'amount': str(amount),
            'currency': str(currency),
            'walletfrom': wallet_from,
            'walletto': wallet_to,
        })

    # Allow you to request a withdrawal from one of your wallet.
    def withdraw(self,
                 w_type: str,
                 wallet: str,
                 amount: float,
                 address: str):
        return self.post('withdraw', json={
            'withdraw_type': str(w_type),
            'walletselected': wallet,
            'amount': str(amount),
            'address': address,
        })

    # ORDERS ------------------------------------------------------------------
    # Submit a new order.
    def place_order(self,
                    amount: float,
                    price: float,
                    side: str,
                    ord_type: str,
                    symbol: str,
                    params: dict=None):
        payload = {
            'symbol': str(symbol),
            'amount': str(amount),
            'price': str(price),
            'side': str(side),
            'type': str(ord_type),
            'ocoorder': False,
        }
        payload.update(params or {})
        return self.post('order/new', payload)

    # Submit a new order.
    def place_oco_order(self,
                        amount: float,
                        price: float,
                        side: str,
                        ord_type: str,
                        symbol: str,
                        buy_price_oco: float,
                        sell_price_oco: float):
        return self.place_order(amount, price, side, ord_type, symbol, params={
            'ocoorder': True,
            'buy_price_oco': str(buy_price_oco),
            'sell_price_oco': str(sell_price_oco),
        })

    # Cancel an order.
    def delete_order(self, order_id: int):
        return self.post('order/cancel', json={'order_id': order_id})

    # Cancel all orders.
    def delete_all_order(self):
        return self.post('order/cancel/all')

    # Get the status of an order. Is it active? Was it cancelled?
    # To what extent has it been executed? etc.
    def status_order(self, order_id: int):
        return self.post('order/status', json={'order_id': order_id})

    # View your active orders.
    def active_orders(self):
        return self.post('orders')

    # View your latest inactive orders.
    # Limited to last 3 days and 1 request per minute.
    def orders_history(self, limit: int):
        return self.post('order/hist', json={'limit': limit})

    # POSITIONS ---------------------------------------------------------------
    # View your active positions.
    def active_positions(self):
        return self.post('positions')

    # Claim a position.
    def claim_position(self, position_id: int):
        return self.post('position/claim', json={'position_id': position_id})

    # HISTORICAL DATA ---------------------------------------------------------
    # View all of your balance ledger entries.
    def balance_history(self,
                        currency: str,
                        since: float=None,
                        until: float=None,
                        limit: int=None,
                        wallet: str=None):
        return self.post('history', json={
            'currency': str(currency),
            'since': since,
            'until': until,
            'limit': limit,
            'wallet': wallet,
        })

    # View your past deposits/withdrawals.
    def movements(self,
                  currency: str,
                  method: str=None,
                  since: float=None,
                  until: float=None,
                  limit: int=None):
        return self.post('history/movements', json={
            'currency': str(currency),
            'method': method,
            'since': since,
            'until': until,
            'limit': limit,
        })

    # View your past trades.
    def past_trades(self,
                    symbol: str,
                    timestamp: float=None,
                    until: float=None,
                    limit_trades: int=None,
                    reverse: bool=None):
        return self.post('mytrades', json={
            'symbol': str(symbol),
            'timestamp': timestamp,
            'until': until,
            'limit_trades': limit_trades,
            'reverse': reverse if reverse is not None else None,
        })

    # MARGIN FUNDING ----------------------------------------------------------
    # Submit a new Offer.
    def place_offer(self,
                    currency: str,
                    amount: float,
                    rate: float,
                    period: int,
                    direction: str):
        return self.post('offer/new', json={
            'currency': str(currency),
            'amount': str(amount),
            'rate': str(rate),
            'period': period,
            'direction': str(direction),
        })

    # Cancel an offer.
    def cancel_offer(self, offer_id: int):
        return self.post('offer/cancel', json={'offer_id': offer_id})

    # Get the status of an offer. Is it active? Was it cancelled?
    # To what extent has it been executed? etc.
    def status_offer(self, offer_id: int):
        return self.post('offer/status', json={'offer_id': offer_id})

    # View your active offers.
    def active_offers(self):
        return self.post('offers')
